/*
 * Generate spherical harmonics plots
 * Need to include Complex.js, WebGLSetup.js
 */

window.addEventListener("load", windowLoadHandler);

function windowLoadHandler() {
    // Get HTML elements
    var glCanvas  = document.getElementById("glCanvas");
    
    var gl = glCanvas.getContext("webgl", {preserveDrawingBuffer: true});
    gl.clearColor(0.0, 0.0, 0.0, 1.0);
    gl.enable(gl.DEPTH_TEST);
    
    // Setup shader program
    var shaderProgram = gl.createProgram();
    WebGLSetup.setupShaders(gl, shaderProgram);
    
    // Setup light
    var lightDirection = [0.0, 1.0, 2.0];
    var lightAmbient   = [0.3, 0.3, 0.3];
    var lightDiffuse   = [0.5, 0.5, 0.5];
    var lightSpecular  = [0.8, 0.8, 0.8];
    var shininess      = 100.0;
    WebGLSetup.setupLight(gl, shaderProgram, lightDirection, lightAmbient, lightDiffuse, lightSpecular, shininess);
    
    // Setup grid buffers
    var gridVertexBuffer = gl.createBuffer();
    var gridNormalBuffer = gl.createBuffer();
    var gridColorBuffer  = gl.createBuffer();
    var gridIndexBuffer  = gl.createBuffer();
    setupGridBuffers(gl, gridVertexBuffer, gridNormalBuffer, gridColorBuffer, gridIndexBuffer);
    
    // Setup spherical harmonics buffers
    var YlmVertexBuffer = gl.createBuffer();
    var YlmNormalBuffer = gl.createBuffer();
    var YlmColorBuffer  = gl.createBuffer();
    var YlmIndexBuffer  = gl.createBuffer();
    setupYlmBuffers(gl, YlmVertexBuffer, YlmNormalBuffer, YlmColorBuffer, YlmIndexBuffer);
    
    // Add event listener
    window.addEventListener("resize", resize, false);
    glCanvas.addEventListener("mousedown", mouseDownHandler, false);
    glCanvas.addEventListener("mousemove", mouseMoveHandler, false);
    glCanvas.addEventListener("mouseup", mouseUpHandler, false);
    glCanvas.addEventListener("mouseout", mouseOutHandler, false);
    glCanvas.addEventListener("mousewheel", mouseWheelHandler, false);
    glCanvas.addEventListener("DOMMouseScroll", mouseWheelHandler, false);
    
    // Event handlers
    // Transformations
    var transMatrix = mat4.create();
    var rotatMatrix = mat4.create();
    mat4.identity(transMatrix);
    mat4.identity(rotatMatrix);
    mat4.translate(transMatrix, [0.0, 0.0, -2.5]);
    mat4.rotate(rotatMatrix, -1.2, [1, 0, 0]);
    mat4.rotate(rotatMatrix, Math.PI*5/4, [0, 0, 1]);
    var oldX = 0, oldY = 0;
    var newX = 0, newY = 0;
    var mouseIsDown = false;
    function mouseDownHandler(evt) {
        evt.preventDefault();
        mouseIsDown = true;
        oldX = evt.clientX;
        oldY = evt.clientY;
    }
    function mouseUpHandler() {mouseIsDown = false;}
    function mouseOutHandler() {mouseIsDown = false;}
    function mouseMoveHandler(evt) {
        if (mouseIsDown == true) {
            newX = evt.clientX;
            newY = evt.clientY;
            var dX = newX - oldX;
            var dY = newY - oldY;
            var newRotatMatrix = mat4.create();
            mat4.identity(newRotatMatrix);
            mat4.rotate(newRotatMatrix, dX/200, [0, 1, 0]);
            mat4.rotate(newRotatMatrix, dY/200, [1, 0, 0]);
            mat4.multiply(newRotatMatrix, rotatMatrix, rotatMatrix);
            oldX = newX;
            oldY = newY;
            refresh();
        }
    }
    function mouseWheelHandler(evt) {
        evt.preventDefault();
        var ds = (evt.wheelDelta>0 || evt.detail<0) ? -0.05*transMatrix[14] : 0.05*transMatrix[14];
        if ((ds>0 && ds>0.02) || (ds<0 && ds>-1)) {
            var newTransMatrix = mat4.create();
            mat4.identity(newTransMatrix);
            mat4.translate(newTransMatrix, [0, 0, ds]);
            mat4.multiply(newTransMatrix, transMatrix, transMatrix);
            refresh();
        }
    }
    
    // Window resize
    function resize() {
        // Adjust width and height
        var canvasWidth  = window.innerWidth - 400;
        var canvasHeight = window.innerHeight - 35;
        if (canvasWidth > 0 && canvasHeight > 0) {
            glCanvas.width  = canvasWidth;
            glCanvas.height = canvasHeight;
        }
        gl.viewportWidth  = canvasWidth;
        gl.viewportHeight = canvasHeight;
        refresh();
    }
    resize();
    
    // Re-draw all objects
    function refresh(){
        WebGLSetup.setupView(gl, shaderProgram, transMatrix, rotatMatrix);
        WebGLSetup.drawObject(gl, shaderProgram, gridVertexBuffer, gridNormalBuffer, gridColorBuffer, gridIndexBuffer, gl.LINES, false);
        WebGLSetup.drawObject(gl, shaderProgram, YlmVertexBuffer, YlmNormalBuffer, YlmColorBuffer, YlmIndexBuffer, gl.LINES, false);
    }
}

function setupGridBuffers(gl, vertexBuffer, normalBuffer, colorBuffer, indexBuffer) {
    var vertices = new Array();
    for (var i=0; i<21; i++) {
        vertices.push(-1+0.1*i);
        vertices.push(-1);
        vertices.push(0);
    }
    for (var i=0; i<21; i++) {
        vertices.push(-1+0.1*i);
        vertices.push(1);
        vertices.push(0);
    }
    for (var i=0; i<21; i++) {
        vertices.push(-1);
        vertices.push(-1+0.1*i);
        vertices.push(0);
    }
    for (var i=0; i<21; i++) {
        vertices.push(1);
        vertices.push(-1+0.1*i);
        vertices.push(0);
    }
    vertices.push(0);
    vertices.push(0);
    vertices.push(-1);
    vertices.push(0);
    vertices.push(0);
    vertices.push(1);
    
    var normals = new Array();
    for (var i=0; i<86; i++) {
        normals.push(0.0);
        normals.push(0.0);
        normals.push(1.0);
    }
    
    var colors = new Array();    
    for (var i=0; i<86; i++) {
        if (i!=10 && i!=31 && i!=52 && i!=73 && i!=84 && i!=85) {
            colors.push(0.4);
            colors.push(0.4);
            colors.push(0.4);
        }
        else if (i==10 || i==31) {
            colors.push(0.0);
            colors.push(1.0);
            colors.push(0.0);
        }
        else if (i==52 || i==73) {
            colors.push(1.0);
            colors.push(0.0);
            colors.push(0.0);
        }
        else {
            colors.push(0.0);
            colors.push(0.0);
            colors.push(1.0);
        }
        colors.push(1.0);
    }
    
    var indices = [
        0, 21, 1, 22, 2, 23, 3 ,24, 4, 25, 5, 26, 6, 27, 7, 28, 8, 29, 9, 30, 10, 31,
        11, 32, 12, 33, 13, 34, 14, 35, 15, 36, 16, 37, 17, 38, 18, 39, 19, 40, 20, 41,
        42, 63, 43, 64, 44, 65, 45, 66, 46, 67, 47, 68, 48, 69, 49, 70, 50, 71, 51, 72, 52, 73,
        53, 74, 54, 75, 55, 76, 56, 77, 57, 78, 58, 79, 59, 80, 60, 81, 61, 82, 62, 83,
        84, 85
    ];
    
    WebGLSetup.setupBuffers(gl, vertices, normals, colors, indices, vertexBuffer, normalBuffer, colorBuffer, indexBuffer);
}

// Create spherical harmonics and setup buffers
function setupYlmBuffers(gl, vertexBuffer, normalBuffer, colorBuffer, indexBuffer) {
    var latN = 20;
    var lonN = 20;
    
    var vertices = new Array();
    var normals  = new Array();
    var colors   = new Array();
    var indices  = new Array();
    
    // Vertices and colors
    for (var i=0; i<=latN; i++) {
        var theta = i / latN * Math.PI;
        for (var j=0; j<=lonN; j++) {
            var phi = j / lonN * 2 * Math.PI;
            
            // Compute spherical harmonics
            
            
            
            
            
            
            
            // Represent result on r component
            var r = 0.5;
            
            // Transform (r,theta,phi) to (x,y,z)
            var x = r*Math.sin(theta)*Math.cos(phi);
            var y = r*Math.sin(theta)*Math.sin(phi);
            var z = r*Math.cos(theta);
            
            vertices.push(x);
            vertices.push(y);
            vertices.push(z);
            
            colors.push(1);
            colors.push(0);
            colors.push(0);
            colors.push(1);
            
            /*
            // Compute phase color
            var phase = 0;
            if (phase>=0 && phase<Math.PI/3) {
                var rat = phase/(Math.PI/3);
                colors.push(1);
                colors.push(rat);
                colors.push(0);
                colors.push(1);
            } else if (phase>=Math.PI/3 && phase<Math.PI*2/3) {
                var rat = (phase-Math.PI/3)/(Math.PI/3);
                colors.push(1-rat);
                colors.push(1);
                colors.push(0);
                colors.push(1);
            } else if (phase>=Math.PI*2/3 && phase<Math.PI) {
                var rat = (phase-Math.PI*2/3)/(Math.PI/3);
                colors.push(0);
                colors.push(1);
                colors.push(rat);
                colors.push(1);
            } else if (phase>=Math.PI && phase<Math.PI*4/3) {
                var rat = (phase-Math.PI)/(Math.PI/3);
                colors.push(0);
                colors.push(1-rat);
                colors.push(1);
                colors.push(1);
            } else if (phase>=Math.PI*4/3 && phase<Math.PI*5/3) {
                var rat = (phase-Math.PI*4/3)/(Math.PI/3);
                colors.push(rat);
                colors.push(0);
                colors.push(1);
                colors.push(1);
            } else {
                var rat = (phase-Math.PI*5/3)/(Math.PI/3);
                colors.push(1);
                colors.push(0);
                colors.push(1-rat);
                colors.push(1);
            }
            */
        }
    }
    
    // Normals
    for (var j=0; j<=lonN; j++) {
        // North pole
        normals.push(0);
        normals.push(0);
        normals.push(0);
    }
    for (var i=1; i<latN; i++) {
        // In between
        var theta    = i / latN * Math.PI;
        var dTheta   = 1 / latN * Math.PI;
        var sinTheta = Math.sin(theta);
        var cosTheta = Math.cos(theta);
        for (var j=0; j<=lonN; j++) {
            var phi    = j / lonN * 2 * Math.PI;
            var dPhi   = 1 / lonN * 2 * Math.PI;
            var sinPhi = Math.sin(phi);
            var cosPhi = Math.cos(phi);
            
            var me    = 3*(i*(lonN+1)+j);
            var up    = 3*((i-1)*(lonN+1)+j);
            var down  = 3*((i+1)*(lonN+1)+j);
            var left  = j==0 ? 3*(i*(lonN+1)+lonN-1) : 3*(i*(lonN+1)+j-1);
            var right = j==lonN ? 3*(i*(lonN+1)+1) : 3*(i*(lonN+1)+j+1);
            
            var Yme    = vertices[me]*vertices[me] + vertices[me+1]*vertices[me+1] + vertices[me+2]*vertices[me+2];
            var Yup    = vertices[up]*vertices[up] + vertices[up+1]*vertices[up+1] + vertices[up+2]*vertices[up+2];
            var Ydown  = vertices[down]*vertices[down] + vertices[down+1]*vertices[down+1] + vertices[down+2]*vertices[down+2];
            var Yleft  = vertices[left]*vertices[left] + vertices[left+1]*vertices[left+1] + vertices[left+2]*vertices[left+2];
            var Yright = vertices[right]*vertices[right] + vertices[right+1]*vertices[right+1] + vertices[right+2]*vertices[right+2];
            
            var rCmpnt     = -1/Yme;
            var thetaCmpnt = 1/(Yme*Yme) * (Ydown-Yup)/(2*dTheta);
            var phiCmpnt   = 1/(Yme*Yme*sinTheta) * (Yright-Yleft)/(2*dPhi);
            
            var xCmpnt = rCmpnt*sinTheta*cosPhi + thetaCmpnt*cosTheta*cosPhi - phiCmpnt*sinPhi;
            var yCmpnt = rCmpnt*sinTheta*sinPhi + thetaCmpnt*cosTheta*sinPhi + phiCmpnt*cosPhi;
            var zCmpnt = rCmpnt*cosTheta - thetaCmpnt*sinTheta;
            
            if (Math.abs(Yme) < 1e-8) {
                // Avoid division by zero
                normals.push(0);
                normals.push(0);
                normals.push(1);
            } else {
                normals.push(-xCmpnt);
                normals.push(-yCmpnt);
                normals.push(-zCmpnt);
            }
        }
    }
    for (var j=0; j<=lonN; j++) {
        // South pole
        normals.push(0);
        normals.push(0);
        normals.push(0);
    }
    
    // Replacing north pole and south pole by taking average from neighbors
    var northNormal = [0, 0, 0];
    var southNormal = [0, 0, 0];
    for (var j=0; j<lonN; j++) {
        for (var k=0; k<3; k++) {
            northNormal[k] = northNormal[k] + normals[3*(1*(lonN+1)+j)+k];
            southNormal[k] = southNormal[k] + normals[3*((latN-1)*(lonN+1)+j)+k];
        }
    }
    for (var k=0; k<3; k++) {
        northNormal[k] = northNormal[k] / lonN;
        southNormal[k] = southNormal[k] / lonN;
    }
    for (var j=0; j<=lonN; j++) {
        normals[3*(0*(lonN+1)+j)]      = northNormal[0];
        normals[3*(0*(lonN+1)+j)+1]    = northNormal[1];
        normals[3*(0*(lonN+1)+j)+2]    = northNormal[2];
        normals[3*(latN*(lonN+1)+j)]   = southNormal[0];
        normals[3*(latN*(lonN+1)+j)+1] = southNormal[1];
        normals[3*(latN*(lonN+1)+j)+2] = southNormal[2];
    }
    
    // Indices for line connection
    for (var i=0; i<=latN; i++) {
        for (var j=0; j<=lonN; j++) {
            var first  = i*lonN + j;
            var second = first + 1;
            indices.push(first);
            indices.push(second);
        }
    }
    for (var j=0; j<lonN; j++) {
        for (var i=0; i<=latN; i++) {
            var first  = i*lonN + j;
            var second = first + lonN + 1;
            indices.push(first);
            indices.push(second);
        }
    }
    
    /*
    // Indices for triangle filling
    for (var i=0; i<latN; i++) {
        for (var j=0; j<lonN; j++) {
            var first  = i*(lonN + 1) + j;
            var second = first + lonN + 1;
            indices.push(first);
            indices.push(second);
            indices.push(first + 1);
            indices.push(second);
            indices.push(second + 1);
            indices.push(first + 1);
        }
    }
    */
    
    WebGLSetup.setupBuffers(gl, vertices, normals, colors, indices, vertexBuffer, normalBuffer, colorBuffer, indexBuffer);
}

// Compute factorial
function factorial(n) {
    var ans = 1;
    for (var i=2; i<=n; i++) {
        ans *= i;
    }
    return ans;
}

// Legendre polynomial Plm(x)
function lgndr(l, m, x) {
    if (m<0 || m>l || Math.abs(x)>1.0) {
        console.log("Invalid arguments for Plm.");
        return 0;
    }
    
    // Compute Pmm
    var Pmm = 1.0;
    if (m > 0) {
        var somx2 = Math.sqrt((1.0+x)*(1.0-x));
        var fact  = 1.0;
        for (var i=1; i<=m; i++) {
            Pmm  = Pmm * (-fact*somx2);
            fact = fact + 2.0;
        }
    }
    
    if (l == m) {
        return Pmm;
    } else {
        // Compute Pmp1m
        var Pmp1m = x*(2*m+1)*Pmm;
        if (l == m+1) {
            return Pmp1m;
        } else {
            // Compute Plm, l > m + 1
            var Plm;
            for (var ll=m+2; ll<=l; ll++) {
                Plm = (x*(2*ll-1)*Pmp1m - (ll+m-1)*Pmm) / (ll-m);
                Pmm = Pmp1m;
                Pmp1m = Plm;
            }
            return Plm;
        }
    }
}
